from abc import abstractmethod
from pprint import pprint
from typing import (
    Any,
    List,
    Literal,
    Protocol,
    Tuple,
    TypedDict,
    runtime_checkable,
)
from urllib.parse import quote

from harambe.tracker import FileDataTracker
from harambe.types import URL, Context, Options, Stage, Cookie, LocalStorage
from playwright.async_api import Page

ObservationTrigger = Literal[
    "on_save_data",
    "on_queue_url",
    "on_download",
    "on_paginate",
    "on_save_cookies",
    "on_save_local_storage",
    "on_check_and_solve_captchas",
]


class DownloadMeta(TypedDict):
    url: str
    filename: str


class HTMLMetadata(DownloadMeta):
    html: str
    text: str


@runtime_checkable
class OutputObserver(Protocol):
    @abstractmethod
    async def on_save_data(self, data: dict[str, Any]) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        raise NotImplementedError()

    @abstractmethod
    async def on_paginate(self, next_url: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_save_cookies(self, cookies: List[Cookie]) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_save_local_storage(self, local_storage: List[LocalStorage]) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def on_check_and_solve_captchas(self, page: Page) -> None:
        raise NotImplementedError()


class LoggingObserver(OutputObserver):
    async def on_save_data(self, data: dict[str, Any]) -> None:
        pprint(data, width=240)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        print(f"Enqueuing: {url} with context {context} and options {options}")

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        print(f"Downloading file: {filename}")  # TODO: use logger
        return {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: List[Cookie]) -> None:
        print(f"Cookies saved : {cookies}")

    async def on_save_local_storage(self, local_storage: List[LocalStorage]) -> None:
        print(f"Local Storage saved : {local_storage}")

    async def on_check_and_solve_captchas(self, page: Page) -> None:
        pass


class LocalStorageObserver(OutputObserver):
    def __init__(self, domain: str, stage: Stage):
        self._tracker = FileDataTracker(domain, stage)

    async def on_save_data(self, data: dict[str, Any]) -> None:
        self._tracker.save_data(data)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        self._tracker.save_data({"url": url, "context": context, "options": options})

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> DownloadMeta:
        data: DownloadMeta = {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }
        self._tracker.save_data(data)  # type: ignore
        return data

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: List[Cookie]) -> None:
        self._tracker.save_data({"cookies": cookies})

    async def on_save_local_storage(self, local_storage: List[LocalStorage]) -> None:
        self._tracker.save_data({"local_storage": local_storage})

    async def on_check_and_solve_captchas(self, page: Page) -> None:
        pass


class InMemoryObserver(OutputObserver):
    def __init__(self) -> None:
        self._data: List[dict[str, Any]] = []
        self._urls: List[Tuple[URL, Context, Options]] = []
        self._files: List[Tuple[str, bytes]] = []
        self._cookies: List[Cookie] = []
        self._local_storage: List[LocalStorage] = []

    async def on_save_data(self, data: dict[str, Any]) -> None:
        self._data.append(data)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        self._urls.append((url, context, options))

    async def on_download(
        self, download_url: str, filename: str, content: bytes
    ) -> "DownloadMeta":
        self._files.append((filename, content))
        return {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
        }

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: List[Cookie]) -> None:
        self._cookies.extend(cookies)

    async def on_check_and_solve_captchas(self, page: Page) -> None:
        pass

    async def on_save_local_storage(self, local_storage: List[LocalStorage]) -> None:
        self._local_storage.extend(local_storage)

    @property
    def data(self) -> List[dict[str, Any]]:
        return self._data

    @property
    def urls(self) -> List[Tuple[URL, Context, Options]]:
        return self._urls

    @property
    def files(self) -> List[Tuple[str, bytes]]:
        return self._files

    @property
    def cookies(self) -> List[Cookie]:
        return self._cookies

    @property
    def local_storage(self) -> List[LocalStorage]:
        return self._local_storage



from bs4.element import Tag
from markdownify import MarkdownConverter


class HTMLToTextConverter(MarkdownConverter):
    """
    Custom converter to convert data from HTML to text

    Strip out standard markdown syntax like headings, em, strong, a, etc.
    Include footnotes in brackets
    """

    def convert_sup(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return f"[{text}]"

    def convert_sub(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return f"[{text}]"

    def convert_span(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        if el.get("class") and "sup" in el.get("class"):
            return f"[{text}]"
        if el.get("class") and "sub" in el.get("class"):
            return f"[{text}]"
        return text

    # Remove markdown headings
    def convert_h1(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_p(el, text, convert_as_inline)

    def convert_h2(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_p(el, text, convert_as_inline)

    def convert_h3(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_p(el, text, convert_as_inline)

    def convert_h4(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_p(el, text, convert_as_inline)

    def convert_h5(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_p(el, text, convert_as_inline)

    def convert_h6(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_p(el, text, convert_as_inline)

    # Treat inline elements as spans
    def convert_strong(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_span(el, text, convert_as_inline)

    def convert_em(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_span(el, text, convert_as_inline)

    def convert_a(self, el: Tag, text: str, convert_as_inline: bool) -> str:
        return self.convert_span(el, text, convert_as_inline)



from typing import Any, Optional

from bs4 import BeautifulSoup, Tag
import json

# noinspection PyProtectedMember
from curl_cffi.requests import AsyncSession, HeaderTypes
from harambe.contrib.soup.tracing import Tracer
from harambe.contrib.types import AbstractElementHandle, AbstractPage, Selectable
from harambe.contrib.types import ResponseWithStatus


class SoupElementHandle(AbstractElementHandle, Selectable["SoupElementHandle"]):
    def __init__(self, tag: Tag) -> None:
        self._tag = tag

    @classmethod
    def from_tag(cls, tag: Tag | None) -> Optional["SoupElementHandle"]:
        return cls(tag) if tag else None

    @classmethod
    def from_tags(cls, tag: list[Tag]) -> list["SoupElementHandle"]:
        return [cls(t) for t in tag]

    async def inner_text(self) -> str:
        return self._tag.get_text()

    async def inner_html(self) -> str:
        return self._tag.decode_contents()

    async def text_content(self) -> str:
        return await self.inner_text()

    async def get_attribute(self, name: str) -> str:
        return self._tag.get(name)

    async def query_selector_all(self, selector: str) -> list["SoupElementHandle"]:
        return self.from_tags(self._tag.select(selector))

    async def query_selector(self, selector: str) -> Optional["SoupElementHandle"]:
        return self.from_tag(self._tag.select_one(selector))

    async def click(self) -> None:
        raise NotImplementedError()

    async def wait_for_selector(self, selector: str, **kwargs: Any) -> None:
        pass


class SoupPage(AbstractPage[SoupElementHandle]):
    _soup: BeautifulSoup
    _url: str

    def __init__(
        self,
        session: AsyncSession,
        extra_headers: Optional[HeaderTypes] = None,
        tracer: Tracer = Tracer(),
    ) -> None:
        self._session = session
        self._extra_headers = extra_headers
        self._tracer = tracer
        self._url = "about:blank"

    @property
    def tracing(self) -> Tracer:
        return self._tracer

    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, value: str) -> None:
        self._url = value

    async def goto(self, url: str, **kwargs: Any) -> ResponseWithStatus:
        res = await self._session.get(
            url, headers=self._extra_headers, impersonate="chrome"
        )
        self._tracer.log_request(res)
        self._url = res.url
        self._soup = BeautifulSoup(res.text, "html.parser")

        class SoupResponseWithStatus:
            status: int = res.status_code
            headers: dict[str, str] = res.headers

        return SoupResponseWithStatus()

    async def post(
        self,
        url: str,
        data: dict[str, Any],
        headers: Optional[HeaderTypes] = None,
        **kwargs: Any,
    ) -> Any:
        res = await self._session.post(
            url,
            headers=headers or self._extra_headers,
            data=json.dumps(data),
            impersonate="chrome",
            **kwargs,
        )
        if self._tracer:
            self._tracer.log_request(res)

        self._url = res.url
        content_type = res.headers.get("Content-Type", "")

        if "application/json" in content_type:

            class SoupResponseWithStatus:
                status: int = res.status_code
                body: dict[str, Any] = res.json()

            return SoupResponseWithStatus()

        self._soup = BeautifulSoup(res.text, "html.parser")

        class SoupResponseWithStatus:
            status: int = res.status_code
            body: str = res.text

        return SoupResponseWithStatus()

    async def query_selector_all(self, selector: str) -> list[SoupElementHandle]:
        return SoupElementHandle.from_tags(self._soup.select(selector))

    async def query_selector(self, selector: str) -> SoupElementHandle | None:
        return SoupElementHandle.from_tag(self._soup.select_one(selector))

    def locator(self, selector: str) -> "SoupLocator":
        return SoupLocator(selector, self)

    async def inner_text(self, selector: str) -> str | None:
        if el := await self.query_selector(selector):
            return await el.inner_text()

        return None

    async def wait_for_timeout(self, timeout: int) -> None:
        pass

    async def content(self) -> str:
        return str(self._soup)

    async def text_content(self, selector: str, **kwargs: Any) -> str | None:
        if el := await self.query_selector(selector):
            return await el.text_content()

        return None

    async def wait_for_selector(self, selector: str, **kwargs: Any) -> None:
        pass

    async def set_extra_http_headers(self, headers: dict[str, str]) -> None:
        self._extra_headers = headers

    async def set_default_timeout(self, timeout: float) -> None:
        pass

    async def wait_for_load_state(self, **kwargs: Any) -> None:
        pass

    async def title(self) -> str | None:
        title = self._soup.title
        if title:
            return title.string
        return None


class SoupLocator:
    def __init__(self, selector: str, page: SoupPage) -> None:
        self._selector = selector
        self._page = page

    async def all(self) -> list[SoupElementHandle]:
        return await self._page.query_selector_all(self._selector)



import pytest

from harambe.observer import InMemoryObserver


@pytest.mark.asyncio
async def in_memory_on_save_data():
    observer = InMemoryObserver()

    await observer.on_save_data({"foo": "bar"})
    await observer.on_save_data({"baz": "qux"})

    assert observer.data == [{"foo": "bar"}, {"baz": "qux"}]


@pytest.mark.asyncio
async def in_memory_on_queue_url():
    observer = InMemoryObserver()

    await observer.on_queue_url("https://example.com", {"foo": "bar"}, {"test": "test"})
    await observer.on_queue_url(
        "https://example.org", {"baz": "qux"}, {"other": "other"}
    )

    assert observer.urls == [
        ("https://example.com", {"foo": "bar"}, {"test": "test"}),
        ("https://example.org", {"baz": "qux"}, {"other": "other"}),
    ]



from harambe_core import Schema

from .core import PAGE_PDF_FILENAME, SDK, AsyncScraper
from .types import AsyncScraperType, ScrapeResult
from .utils import PlaywrightUtils

__all__ = [
    "ScrapeResult",
    "SDK",
    "PlaywrightUtils",
    "AsyncScraperType",
    "AsyncScraper",
    "PAGE_PDF_FILENAME",
    "Schema",
]



import pytest
from harambe_core.errors import CaptchaError, HarambeException


def simulate_captcha_scenario():
    # Simulating a CAPTCHA hit scenario
    raise CaptchaError()


def test_captcha_exception_is_harambe_exception():
    # Testing if HarambeException is raised when expected
    with pytest.raises(HarambeException):
        simulate_captcha_scenario()



import pytest

from harambe.proxy import proxy_from_url


@pytest.mark.parametrize(
    "url, expected",
    [
        (
            "http://user:pass@hostname:8080",
            {"server": "hostname:8080", "username": "user", "password": "pass"},
        ),
        (
            "user:pass@hostname:8080",
            {"server": "hostname:8080", "username": "user", "password": "pass"},
        ),
        (
            "http://user:pass@hostname",
            {"server": "hostname", "username": "user", "password": "pass"},
        ),
        (
            "user:pass@hostname",
            {"server": "hostname", "username": "user", "password": "pass"},
        ),
    ],
)
def test_proxy_from_url_valid(url, expected):
    assert proxy_from_url(url) == expected


@pytest.mark.parametrize(
    "url",
    [
        "http://hostname:8080",
        "hostname:8080",
        "http://user@hostname:8080",
        "user@hostname:8080",
        "http://hostname",
        "hostname",
    ],
)
def test_proxy_from_url_invalid(url):
    with pytest.raises(ValueError):
        proxy_from_url(url)



import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, List

from harambe.meta import url_to_netloc


class DataTracker(ABC):
    @abstractmethod
    async def save_data(self, new_data: List[dict[str, Any]]) -> None:
        """Append data for a domain and stage to a JSON file"""
        raise NotImplementedError()

    @abstractmethod
    def load_data(self, url: str, stage: str) -> List[dict[str, Any]]:
        """Load data for a domain and stage from a JSON file"""
        raise NotImplementedError()

    @abstractmethod
    def visit(self, url: str) -> None:
        """Mark the url as visited and save to the domain file"""
        raise NotImplementedError()

    @abstractmethod
    def has_been_visited(self, url: str) -> bool:
        """Check if the url has been visited"""
        raise NotImplementedError()


class StubTracker(DataTracker):
    async def save_data(self, new_data: list[dict[str, Any]]) -> None:
        pass

    def load_data(self, url: str, stage: str) -> list[dict[str, Any]]:
        return []

    def visit(self, url: str) -> None:
        pass

    def has_been_visited(self, url: str) -> bool:
        return False


class FileDataTracker(DataTracker):
    def __init__(self, domain: str, stage: str):
        # Define the storage directory as 'data' one level up from the current script
        self.storage_dir = Path(__file__).resolve().parent.parent / "data"
        # Ensure the storage directory exists
        self.storage_dir.mkdir(parents=True, exist_ok=True)
        self.visited_urls: dict[str, set[str]] = {}
        self.domain = url_to_netloc(domain)
        self.stage = stage

    def get_storage_filepath(self, data_type: str) -> Path:
        """Generate the filename for the given domain and data type"""
        return self.storage_dir / f"{self.domain}_{data_type}.json"

    def _load_visited_urls(self, domain: str) -> set[str]:
        """Load visited urls for a domain from a JSON file"""
        domain_file = self.get_storage_filepath("urls")
        if domain_file.exists():
            with domain_file.open("r") as file:
                return set(json.load(file))
        return set()

    def _save_visited_urls(self, domain: str) -> None:
        """Save visited urls for a domain to a JSON file"""
        domain_file = self.get_storage_filepath("urls")
        with domain_file.open("w") as file:
            json.dump(list(self.visited_urls.get(domain, set())), file, indent=4)

    def save_data(self, *new_data: dict[str, Any]) -> None:  # type: ignore
        """Append data for a domain and stage to a JSON file"""
        domain_file = self.get_storage_filepath(self.stage)

        # Load existing data if file exists
        if domain_file.exists():
            with domain_file.open("r") as file:
                data = json.load(file)
        else:
            data = []

        # Append new data
        data.extend(new_data)

        # Save the updated data back to the JSON file
        with domain_file.open("w") as file:
            json.dump(data, file, indent=4)

    def load_data(self, url: str | None, stage: str | None) -> list[dict[str, Any]]:
        """Load data for a domain and stage from a JSON file"""
        domain_file = self.get_storage_filepath(stage)  # type: ignore
        if domain_file.exists():
            with domain_file.open("r") as file:
                return json.load(file)
        return []

    def visit(self, url: str) -> None:
        """Mark the url as visited and save to the domain file"""
        domain = self.domain
        if domain not in self.visited_urls:
            self.visited_urls[domain] = self._load_visited_urls(domain)
        self.visited_urls[domain].add(url)
        self._save_visited_urls(domain)

    def has_been_visited(self, url: str) -> bool:
        """Check if the url has been visited"""
        domain = url_to_netloc(url)
        if domain not in self.visited_urls:
            self.visited_urls[domain] = self._load_visited_urls(domain)
        return url in self.visited_urls[domain]



from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Awaitable, Callable, Optional, Sequence

from curl_cffi.requests import AsyncSession

from harambe.contrib.soup.impl import SoupPage
from harambe.contrib.soup.tracing import Tracer
from harambe.types import SetCookieParam

Callback = Callable[[Tracer], Awaitable[None]]
PageFactory = Callable[..., Awaitable[SoupPage]]


@asynccontextmanager
async def soup_harness(
    *,
    proxy: str | None = None,
    cookies: Sequence[SetCookieParam] = (),
    headers: dict[str, str] | None = None,
    on_start: Optional[Callback] = None,
    on_end: Optional[Callback] = None,
    **__: Any,
) -> AsyncGenerator[PageFactory, None]:
    async with AsyncSession(proxy=proxy, impersonate="chrome", verify=False) as s:
        for c in cookies:
            s.cookies.set(
                name=c["name"],
                value=c["value"],
                domain=c.get("domain", ""),
                path=c.get("path", "/"),
                secure=c.get("secure", False),
            )

        tracer = Tracer()

        async def factory(*_: Any, **__: Any) -> SoupPage:
            page = SoupPage(s, tracer=tracer)
            if headers:
                await page.set_extra_http_headers(headers)
            return page

        try:
            if on_start:
                await on_start(tracer)

            yield factory
        finally:
            if on_end:
                await on_end(tracer)



import json

import pytest

from harambe.contrib import playwright_harness


@pytest.fixture
async def async_page():
    async with playwright_harness(
        headless=True, enable_clipboard=True, browser_type="chromium"
    ) as page_factory:
        yield await page_factory()


@pytest.mark.asyncio
async def test_navigator_webdriver(async_page):
    await async_page.goto("https://kaliiiiiiiiii.github.io/brotector/?crash=false")

    # Site contains a button to save the results as JSON
    await async_page.get_by_role("button", name="copy as JSON").click()
    clipboard = await async_page.evaluate("navigator.clipboard.readText()")
    stealth_results = json.loads(clipboard)

    # There are many different checks on the site. But we're focusing our attention on specific ones that we already fix
    # TODO: Add more checks as needed
    detections_to_avoid = [
        "navigator.webdriver",
    ]

    assert not any(
        result["detection"] in detections_to_avoid for result in stealth_results
    ), "A detection that should be avoided was found in the results"


@pytest.mark.asyncio
async def test_user_agent(async_page):
    user_agent = await async_page.evaluate("navigator.userAgent")
    assert "headless" not in user_agent.lower(), (
        "User agent should not contain 'headless'"
    )


@pytest.mark.asyncio
async def test_window_chrome(async_page):
    has_chrome = await async_page.evaluate("!!window.chrome")
    assert has_chrome, "window.chrome should be present in non-headless Chrome"


@pytest.mark.asyncio
async def test_plugins(async_page):
    plugins_length = await async_page.evaluate("navigator.plugins.length")
    assert plugins_length > 0, "Non-headless browsers should have at least one plugin"


@pytest.mark.asyncio
async def test_app_version(async_page):
    app_version = await async_page.evaluate("navigator.appVersion")
    assert "headless" not in app_version.lower(), (
        "App version should not contain 'headless'"
    )


@pytest.mark.asyncio
async def test_notification_permissions(async_page):
    permission_status = await async_page.evaluate("""
    async () => {
        const permissionStatus = await navigator.permissions.query({name: 'notifications'});
        return {
            notificationPermission: Notification.permission,
            permissionState: permissionStatus.state
        };
    }
    """)

    assert not (
        permission_status["notificationPermission"] == "denied"
        and permission_status["permissionState"] == "prompt"
    ), "Notification permissions should not be in an inconsistent state"


@pytest.mark.asyncio
async def test_connection_rtt(async_page):
    connection_rtt = await async_page.evaluate("""
    () => {
        const connection = navigator.connection;
        return connection ? connection.rtt : undefined;
    }
    """)

    assert connection_rtt is not None and connection_rtt != 0, (
        "Connection RTT should exist and not be zero in non-headless browsers"
    )



import functools
import re
from typing import Callable, List, Optional, ParamSpec, TypeVar, Union
from urllib.parse import urljoin

from playwright.async_api import Locator, Page

PageLocator = Union[Locator, Page]


class PlaywrightUtils:
    @staticmethod
    async def parse_by_regex(page: PageLocator, selector: str) -> str:
        return await parse_by_regex(page, selector)  # type: ignore

    @staticmethod
    async def parse_attr(page: PageLocator, selector: str) -> str:
        return await parse_attr(page, selector)

    @staticmethod
    async def get_texts(page: PageLocator, selector: str) -> List[str]:
        return await get_texts(page, selector)

    @staticmethod
    async def get_text(page: PageLocator, selector: str) -> str:
        return await get_text(page, selector)

    @staticmethod
    async def get_attrs(page: PageLocator, selector: str, attr: str) -> List[str]:
        return await get_attrs(page, selector, attr)

    @staticmethod
    async def get_attr(page: PageLocator, selector: str, attr: str) -> str:
        return await get_attr(page, selector, attr)

    @staticmethod
    async def get_links(page: PageLocator, selector: str) -> List[str]:
        return await get_links(page, selector)

    @staticmethod
    async def get_link(page: PageLocator, selector: str) -> str:
        return await get_link(page, selector)


async def parse_by_regex(page: PageLocator, selector: str) -> Optional[str]:
    page_source = await (
        page.evaluate("(element) => element.outerHTML")
        if isinstance(page, Locator)
        else page.content()
    )

    data = re.findall(selector, page_source)
    return data[0] if data else None


async def parse_attr(page: PageLocator, selector: str) -> str:
    xpath, attr = selector.split(r"/@")
    data = await get_attr(page, xpath, attr)
    if data.startswith("/"):
        data = urljoin(page.url, data)  # type: ignore
    return data


async def get_texts(page: PageLocator, selector: str) -> List[str]:
    elements = await page.locator(selector).all()
    data = [await el.text_content() for el in elements]
    return [d.strip() for d in data if d]


async def get_text(page: PageLocator, selector: str) -> str:
    data = await get_texts(page, selector)
    return data[0] if data else ""


async def get_attrs(page: PageLocator, selector: str, attr: str) -> List[str]:
    elements = await page.locator(selector).all()
    data = [await el.get_attribute(attr) for el in elements]
    return [d.strip() for d in data if d]


async def get_attr(page: PageLocator, selector: str, attr: str) -> str:
    data = await get_attrs(page, selector, attr)
    return data[0] if data else ""


# TODO: edge case where page is a Locator and url is not set
async def get_links(page: PageLocator, selector: str) -> List[str]:
    urls = await get_attrs(page, selector, attr="href")
    if isinstance(page, Locator):
        setattr(page, "url", "")
    return [urljoin(page.url, url) for url in urls if url]  # type: ignore


async def get_link(page: PageLocator, selector: str) -> str:
    data = await get_links(page, selector)
    return data[0] if data else ""


P = ParamSpec("P")
T = TypeVar("T")


def swallow_exceptions(func: Callable[P, T]) -> Callable[P, Optional[T]]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            print(f"Exception swallowed in {func.__name__}: {e}")
            return None

    return wrapper



import asyncio
from typing import Callable, Awaitable, cast

import ua_generator

UserAgentFactory = str | Callable[[], str | Awaitable[str]]


def random_user_agent() -> str:
    ua = ua_generator.generate(device="desktop", browser=("chrome", "edge"))
    return ua.headers.get()["user-agent"]


async def compute_user_agent(factory: UserAgentFactory) -> str:
    if not callable(factory):
        return factory

    result = factory()
    if asyncio.iscoroutine(result):
        return await result

    return cast(str, result)



import ast
from pathlib import Path
from typing import List, TypedDict, cast
from urllib.parse import urlparse


class DecoratedScraper(TypedDict):
    file_path: str
    function_name: str
    domain: str
    stage: str
    package: str


def is_sdk_scraper_decorator(node: ast.expr) -> bool:
    return (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Attribute)
        and node.func.attr == "scraper"
        and isinstance(node.func.value, ast.Name)
        and node.func.value.id == "SDK"
    )


def find_decorated_scrapers(file_path: Path) -> List[DecoratedScraper]:
    with open(file_path, "r") as source:
        node = ast.parse(source.read())

    decorated_methods: List[DecoratedScraper] = []
    for func in [n for n in node.body if isinstance(n, ast.AsyncFunctionDef)]:
        for decorator in func.decorator_list:
            if is_sdk_scraper_decorator(decorator):
                decorator = cast(ast.Call, decorator)
                kws = {kw.arg: kw.value.s for kw in decorator.keywords}  # type: ignore
                domain = kws["domain"]

                decorated_methods.append(
                    {
                        "file_path": str(file_path),
                        "function_name": func.name,
                        "domain": domain,
                        "stage": kws["stage"],
                        "package": url_to_package(domain),
                    }
                )

    return decorated_methods


def walk_package_for_decorators(path: Path) -> List[DecoratedScraper]:
    files = (
        [path]
        if path.is_file()
        else [p for p in path.rglob("*.py") if not p.name.startswith("_")]
    )

    decorated_methods_in_package = []
    for file in files:
        decorated_methods_in_package.extend(find_decorated_scrapers(file))

    return decorated_methods_in_package


def url_to_netloc(url: str) -> str:
    if not url.startswith("http://") and not url.startswith("https://"):
        url = "https://" + url

    domain = urlparse(url).netloc
    domain = domain.split(":")[0]
    if domain.startswith("www."):
        domain = domain[4:]

    return domain


def url_to_package(url: str) -> str:
    domain = url_to_netloc(url)

    parts = domain.split(".")
    reversed_parts = parts[::-1]

    package_name = ".".join(reversed_parts)
    return package_name



import re
import base64
import time
from abc import ABC
from typing import Any, Literal, Self

from playwright.async_api import Page, Route

ResourceType = Literal[
    "document",
    "stylesheet",
    "image",
    "media",
    "font",
    "script",
    "texttrack",
    "xhr",
    "fetch",
    "eventsource",
    "websocket",
    "manifest",
    "other",
    "*",
]

FAKE_IMAGE_BYTES = base64.b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII="
)


class AbstractHandler(ABC):
    async def handle(self, route: Route) -> None:
        raise NotImplementedError


class ResourceRequestHandler(AbstractHandler):
    def __init__(
        self,
        page: Page,
        resource_type: ResourceType,
        timeout: int,
        url_pattern: str = "**/*",
    ):
        self.page = page
        self.url_pattern = url_pattern
        self.resource_type = resource_type
        self.timeout = timeout
        self._initial_pages = [p.url for p in page.context.pages]
        self._new_pages: list[str] = []

    async def __aenter__(self) -> Self:
        await self.page.context.route(self.url_pattern, self.handle)
        return self

    async def __aexit__(self, *_: Any, **__: Any) -> None:
        await self.page.context.unroute(self.url_pattern, self.handle)
        await self.page.bring_to_front()
        try:
            new_page = await self._wait_for_new_page()
            self._new_pages.append(new_page.url)
            await new_page.close()
        except TimeoutError:
            raise TimeoutError(
                f"No new page opened within the {self.timeout} ms timeout."
            )

    async def _wait_for_new_page(self) -> Page:
        start_time = time.monotonic()
        timeout_seconds = self.timeout / 1000
        while time.monotonic() - start_time < timeout_seconds:
            for page in self.page.context.pages:
                if page.url not in self._initial_pages:
                    return page
            await self.page.wait_for_timeout(100)
        raise TimeoutError("Timed out waiting for a new page to open.")

    async def handle(self, route: Route) -> None:
        if (
            self.resource_type == "*"
            or self.resource_type in route.request.resource_type
        ):
            await route.fulfill(
                status=200,
                content_type="text/plain",
                body="Intercepted by the handler",
            )
            return

        await route.fallback()

    def captured_url(self) -> str | None:
        if len(self._new_pages) > 1:
            raise ValueError("More than one page matched")

        return self._new_pages[0] if self._new_pages else None


class UnnecessaryResourceHandler:
    async def handle(self, route: Route) -> None:
        resource_type = route.request.resource_type
        url = route.request.url
        if resource_type in ["image", "media"]:
            await route.fulfill(body=FAKE_IMAGE_BYTES, content_type="image/png")
            return
        elif (
            resource_type == "font"
            or re.match(r"^data:(image|audio|video)", url)
            or re.match(r"social-widget|tracking-script|ads", url)
        ):
            await route.abort("blockedbyclient")
            return

        await route.fallback()



import hashlib
import json
from typing import Any, Optional, Iterable, List

from pydantic import BaseModel

from harambe.types import URL, Context, Options, Cookie, LocalStorage


class PageInfo(BaseModel):
    page: int
    total_rows: int
    duplicated_rows: int


class DuplicateHandler:
    def __init__(self) -> None:
        self._saved_data: set[bytes] = set()
        self.current_page: int = 1
        self.page_info_map: dict[int, PageInfo] = {}

    def on_save_data(self, data: dict[str, Any]) -> bool:
        """
        Save data and check if it is duplicated
        :param data: data to be saved
        :return: bool indicating if the data is duplicated, true if it is duplicated
        """

        return self._add_data(data)

    def on_save_cookies(self, cookies: List[Cookie]) -> bool:
        """
        Save cookies and check if they are duplicated
        :param cookies: cookies to be saved
        :return: bool indicating if the cookies are duplicated, true if it is duplicated
        """
        return self._add_data(cookies)

    def on_save_local_storage(self, local_storage: List[LocalStorage]) -> bool:
        """
        Save local storage and check if they are duplicated
        :param local_storage: local storage to be saved
        :return: bool indicating if the local storage is duplicated, true if it is duplicated
        """
        return self._add_data(local_storage)

    def on_queue_url(
        self, url: URL, _: Optional[Context], __: Optional[Options]
    ) -> bool:
        """
        Save url and check if it is duplicated
        :param url: url to be saved
        :param _: context unused
        :param __: options unused
        :return: bool indicating if the url is duplicated, true if it is duplicated
        """
        return self._add_data(url)

    # noinspection PyTypeChecker
    def on_download(self, download_url: str, filename: str, content: bytes) -> bool:
        return self._add_data((download_url, filename))

    def should_continue(self, strict: bool = False) -> bool:
        """
        Check if the current page should be continued to be processed
        :param strict: if strict is True, it will only return True if any rows on the page are duplicated,
        otherwise it will return True if and only iff all rows are duplicated
        :return: bool indicating if pagination should continued
        """

        if strict:
            return self.get_current_page_info().duplicated_rows == 0

        return (
            self.get_current_page_info().total_rows > 0
            and self.get_current_page_info().total_rows
            == self.get_current_page_info().duplicated_rows
        )

    def on_paginate(self, next_url: str) -> bool:
        if self.should_continue():
            raise StopAsyncIteration()

        self.current_page += 1
        return False

    def _add_data(self, data: Any) -> bool:
        """
        :return: bool indicating whether the data is duplicated or not
        """
        self.get_current_page_info().total_rows += 1

        hash_value = self.compute_hash(data)
        if hash_value in self._saved_data:
            self.get_current_page_info().duplicated_rows += 1
            return True
        else:
            self._saved_data.add(hash_value)
            return False

    def get_number_of_pages(self) -> int:
        return self.current_page

    def get_current_page_info(self) -> PageInfo:
        if self.current_page not in self.page_info_map:
            self.page_info_map[self.current_page] = PageInfo(
                page=self.current_page, total_rows=0, duplicated_rows=0
            )

        return self.page_info_map[self.current_page]

    @staticmethod
    def compute_hash(data: Any) -> bytes:
        if isinstance(data, dict):
            data = {k: v for k, v in data.items() if not k.startswith("__")}

        data_str = json.dumps(data, separators=(",", ":"), sort_keys=True)
        return hashlib.md5(data_str.encode()).digest()


class PaginatedList(list[Any]):
    def __init__(self) -> None:
        super().__init__()
        self._handler = DuplicateHandler()

    def append(self, item: Any) -> None:
        if not self._handler.on_save_data(item):
            super().append(item)

    def extend(self, items: Iterable[Any]) -> None:
        for item in items:
            self.append(item)

    def should_continue(self) -> bool:
        return self._handler.should_continue(strict=True)



import asyncio
import inspect
import tempfile
import uuid
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    Callable,
    List,
    Optional,
    Protocol,
    Union,
    Unpack,
    cast,
    Tuple,
)

import aiohttp
from bs4 import BeautifulSoup, Doctype
from harambe.contrib.soup.impl import SoupPage
from harambe.contrib.types import AbstractPage
from harambe.cookie_utils import fix_cookie
from harambe.handlers import (
    ResourceRequestHandler,
    ResourceType,
)
from harambe.html_converter import HTMLConverterType, get_html_converter
from harambe.observer import (
    DownloadMeta,
    HTMLMetadata,
    LocalStorageObserver,
    LoggingObserver,
    ObservationTrigger,
    OutputObserver,
)
from harambe.pagination import DuplicateHandler
from harambe.tracker import FileDataTracker
from harambe.types import (
    URL,
    AsyncScraperType,
    Context,
    HarnessOptions,
    Options,
    ScrapeResult,
    SetupType,
    Stage,
    Cookie,
    LocalStorage,
)
from harambe_core import SchemaParser, Schema
from harambe_core.errors import default_error_callback
from harambe_core.normalize_url import normalize_url
from harambe_core.parser.expression import ExpressionEvaluator
from playwright.async_api import (
    ElementHandle,
    Page,
)
from playwright.async_api import (
    TimeoutError as PlaywrightTimeoutError,
)

from harambe.contrib import WebHarness, playwright_harness


class AsyncScraper(Protocol):
    """
    Protocol that all classed based scrapers should implement.
    Note that scrapers in harambe should be functions, not classes.
    """

    async def scrape(self, sdk: "SDK", url: URL, context: Context) -> None: ...


class SDK:
    """
    A web scraping SDK with a number of useful methods and features for:
    - Saving data and validating it following a specific schema
    - Enqueuing (and automatically formatting) urls
    - De-duplicating data, urls, etc
    - Effectively handling classic web scraping problems like pagination, pdfs, downloads, etc
    """

    def __init__(
        self,
        page: AbstractPage[Any],
        run_id: Optional[str] = None,
        domain: Optional[str] = None,
        stage: Optional[Stage] = None,
        observer: Optional[Union[OutputObserver, List[OutputObserver]]] = None,
        scraper: Optional[AsyncScraperType] = None,
        context: Optional[Context] = None,
        schema: Optional[Schema] = None,
        deduper: Optional[DuplicateHandler] = None,
        evaluator: Optional[ExpressionEvaluator] = None,
    ):
        self.page: Page = page  # type: ignore
        self._id = run_id or uuid.uuid4()
        self._domain = domain
        self._stage = stage
        self._scraper = scraper
        self._context = context or {}
        self._validator = SchemaParser(schema, evaluator) if schema else None
        self._saved_data: set[ScrapeResult] = set()
        self._saved_cookies: List[Cookie] = []
        self._saved_local_storage: List[LocalStorage] = []

        if not observer:
            observer = [LoggingObserver()]

        if not isinstance(observer, list):
            observer = [observer]

        self._observers = observer
        self._deduper = DuplicateHandler()
        self._deduper = deduper if deduper else DuplicateHandler()

    async def save_data(self, *data: ScrapeResult) -> None:
        """
        Save scraped data and validate its type matches the current schema

        :param data: one or more dictionaries of data to save
        :raises SchemaValidationError: If any of the saved data does not match the provided schema
        :example:
            >>> await sdk.save_data({ "title": "example", "description": "another_example" })
        """
        if len(data) == 1 and isinstance(data[0], list):
            raise TypeError(
                "`SDK.save_data` should be called with one dict at a time, not a list of dicts."
            )

        source_url = self.page.url
        for d in data:
            if self._validator is not None:
                d = self._validator.validate(d, base_url=source_url)
            d["__url"] = source_url
            await self._notify_observers("on_save_data", d)

    async def enqueue(
        self,
        *urls: URL | Awaitable[URL],
        context: Optional[Context] = None,
        options: Optional[Options] = None,
    ) -> None:
        """
        Enqueue url(s) to be scraped later.

        :param urls: urls to enqueue
        :param context: additional context to pass to the next run of the next stage/url. Typically just data that is only available on the current page but required in the schema.
        :param options: job level options to pass to the next stage/url

        :example:
            >>> await sdk.enqueue("https://www.test.com")
            >>> await sdk.enqueue("/some-path") # This will automatically be converted into an absolute url
        """
        context = context or {}
        options = options or {}
        context["__url"] = self.page.url

        for url in urls:
            if inspect.isawaitable(url):
                url = await url

            normalized_url = (
                normalize_url(url, self.page.url) if hasattr(self.page, "url") else url
            )
            await self._notify_observers(
                "on_queue_url", normalized_url, context, options
            )

    async def paginate(
        self,
        get_next_page_element: Callable[..., Awaitable[URL | ElementHandle | None]],
        timeout: int = 2000,
    ) -> None:
        """
        SDK method to automatically facilitate paginating a list of elements.
        Simply define a function that should return any of:
            - A direct link to the next page
            - An element with hrefs to the next page
            - An element to click on to get to the next page
        And call `sdk.paginate` at the end of your scrape function. The element will automatically be used to paginate the site and run the scraping code against all pages
        Pagination will conclude once all pages have been reached.

        This method should ALWAYS be used for pagination instead of manual for loops and if statements.

        :param get_next_page_element: the url or ElementHandle of the next page
        :param timeout: milliseconds to sleep for before continuing. Only use if there is no other wait option

        :example:
            >>> async def pager():
            ...     return await page.query_selector("div.pagination > .pager.next")
            ...
            ... await sdk.paginate(pager)
        """
        try:
            next_page = await get_next_page_element()
            if not next_page:
                return

            next_url = ""
            if isinstance(next_page, ElementHandle):
                await next_page.click(timeout=timeout)
                await self.page.wait_for_timeout(timeout)
                next_url = self.page.url

            elif isinstance(next_page, str):
                next_url = next_page
                if next_url.startswith("?"):
                    # TODO: merge query params
                    next_url = self.page.url.split("?")[0] + next_url

                await self.page.goto(normalize_url(next_url, self.page.url))
                await self.page.wait_for_timeout(timeout)

            if next_url:
                await self._notify_observers("on_paginate", next_url)
                if not self._scraper:
                    return
                await self._scraper(
                    self, next_url, self._context
                )  # TODO: eventually fix this to not be recursive
        except PlaywrightTimeoutError as e:
            raise TimeoutError(
                f"{e.args[0]} You may increase the timeout by passing `timeout` in ms to `SDK.paginate`. Alternatively, this may mean that the next page element or URL was not found and pagination is complete."
            ) from e
        except (TimeoutError, AttributeError, StopAsyncIteration):
            return

    async def capture_url(
        self,
        clickable: ElementHandle,
        resource_type: ResourceType = "document",
        timeout: Optional[int] = 10000,
    ) -> URL | None:
        """
        Capture the url of a click event. This will click the element and return the url
        via network request interception. This is useful for capturing urls that are
        generated dynamically (eg: redirects to document downloads).

        :param clickable: the element to click
        :param resource_type: the type of resource to capture
        :param timeout: the time to wait for the new page to open (in ms)
        :return url: the url of the captured resource or None if no match was found
        :raises ValueError: if more than one request matches
        """
        async with ResourceRequestHandler(
            self.page, resource_type=resource_type, timeout=timeout
        ) as handler:
            await clickable.click()

        return handler.captured_url()

    async def capture_download(
        self,
        clickable: ElementHandle,
        override_filename: str | None = None,
        override_url: str | None = None,
        timeout: float | None = None,
    ) -> DownloadMeta:
        """
        Capture a download event that gets triggered by clicking an element. This method will:
         - Handle clicking the element
         - Download the resulting file
         - Apply download handling logic and build a download URL
         - Return a download metadata object

        Use this method to manually download dynamic files or files that can only be downloaded in the current browser session.

        :return DownloadMeta: A typed dict containing the download metadata such as the `url` and `filename`
        """

        async with self.page.expect_download(timeout=timeout) as download_info:
            await clickable.click()
        download = await download_info.value

        # Create a temporary file to save the download
        with tempfile.NamedTemporaryFile() as temp_file:
            await download.save_as(temp_file.name)
            with open(temp_file.name, "rb") as f:
                content = f.read()

        res = await self._notify_observers(
            "on_download",
            override_url if override_url else download.url,
            override_filename if override_filename else download.suggested_filename,
            content,
            check_duplication=False,
        )
        return res[0]

    async def capture_html(
        self,
        selector: str = "html",
        exclude_selectors: List[str] | None = None,
        *,
        soup_transform: Optional[Callable[[BeautifulSoup], None]] = None,
        html_converter_type: HTMLConverterType = "markdown",
    ) -> HTMLMetadata:
        """
        Capture and download the html content of the document or a specific element.
        The returned HTML will be cleaned of any excluded elements and will be wrapped in a proper HTML document structure.

        :param selector: CSS selector of element to capture. Defaults to "html" for the document element.
        :param exclude_selectors: List of CSS selectors for elements to exclude from capture.
        :param soup_transform: A function to transform the BeautifulSoup html prior to saving. Use this to remove aspects of the returned content
        :param html_converter_type: Type of HTML converter to use for the inner text. Defaults to "markdown".
        :return: HTMLMetadata containing the `html` of the element, the formatted `text` of the element, along with the `url` and `filename` of the document
        :raises ValueError: If the specified selector doesn't match any element.
        :example:
            >>> meta = await sdk.capture_html(selector="div.content")
            ... await sdk.save_data({"name": meta["filename"], "text": meta["text"], "download_url": meta["url"]})
        """
        html, text = await self._get_html(
            selector,
            exclude_selectors or [],
            soup_transform or (lambda x: None),
            html_converter_type,
        )

        downloads = await self._notify_observers(
            method="on_download",
            download_url=self.page.url,
            filename=f"{str(uuid.uuid4())}.html",
            content=html.encode("utf-8"),
            check_duplication=False,
        )

        return {
            "url": downloads[0]["url"],
            "filename": downloads[0]["filename"],
            "html": html,
            "text": text,
        }

    async def _get_html(
        self,
        selector: str,
        exclude_selectors: List[str],
        soup_transform: Callable[[BeautifulSoup], None],
        html_converter_type: HTMLConverterType,
    ) -> Tuple[str, str]:
        element = await self.page.query_selector(selector)

        if not element:
            raise ValueError(f"Element not found for selector: {selector}")

        raw_inner_html = await element.inner_html()

        soup = BeautifulSoup(raw_inner_html, "html.parser")
        for selector in exclude_selectors:
            for element_to_remove in soup.select(selector):
                element_to_remove.decompose()

        soup_transform(soup)

        # Ensure HTML doc type is present so the file can be correctly parsed
        if not soup.contents or not isinstance(soup.contents[0], Doctype):
            doctype = Doctype("html")
            soup.insert(0, doctype)

        text = get_html_converter(html_converter_type).convert_soup(soup)

        return str(soup), text

    async def capture_pdf(
        self,
    ) -> DownloadMeta:
        """
        Capture the current page as a pdf and then apply some download handling logic
        from the observer to transform to a usable URL

        :return DownloadMeta: A typed dict containing the download metadata such as the `url` and `filename`
        :example:
            >>> meta = await sdk.capture_pdf()
            ... await sdk.save_data({"file_name": meta["filename"], "download_url": meta["url"]})
        """
        await self.page.wait_for_timeout(
            1000
        )  # Allow for some extra time for the page to load
        pdf_content = await self.page.pdf()
        file_name = PAGE_PDF_FILENAME
        res = await self._notify_observers(
            "on_download", self.page.url, file_name, pdf_content
        )
        return res[0]

    async def save_cookies(
        self,
        override_cookies: Optional[List[Cookie]] = None,
    ) -> None:
        """
        Save the cookies from the current browser context or use the provided cookies.

        This function retrieves all the cookies from the current browser context if none are provided,
        saves them to the SDK instance, and notifies all observers about the action performed.

        :param override_cookies: Optional list of cookie dictionaries to save. If None, cookies are retrieved from the current page context.
        """
        new_cookies = override_cookies or cast(
            List[Cookie], await self.page.context.cookies()
        )
        new_cookies = [fix_cookie(cookie) for cookie in new_cookies]

        self._saved_cookies = self._saved_cookies + new_cookies

        await self._notify_observers("on_save_cookies", new_cookies)

    async def save_local_storage(
        self,
        override_local_storage: Optional[dict[str, str]] = None,
        override_domain: str | None = None,
        override_path: str | None = None,
    ) -> None:
        """
        Save the local storage from the current browser context or provided local storage.

        This function retrieves all the local storage data from the current page context if none is provided,
        updates the SDK instance with new or updated values, and notifies all observers about the action performed.

        :param override_local_storage: Optional list of local storage items (key-value pairs) to save. If None, local storage is retrieved from the current page context.
        :param override_domain: Optional domain to use for the local storage
        :param override_path: Optional path to use for the local storage
        """
        new_browser_local_storage = override_local_storage or await self.page.evaluate(
            "() => localStorage"
        )

        domain = override_domain or self._domain
        if not domain:
            raise RuntimeError("No domain provided for local storage")

        new_local_storage = [
            LocalStorage(
                domain=domain,
                path=override_path or "/",
                key=key,
                value=new_browser_local_storage[key],
            )
            for key in new_browser_local_storage.keys()
        ]

        self._saved_local_storage = self._saved_local_storage + new_local_storage

        await self._notify_observers("on_save_local_storage", new_local_storage)

    async def solve_captchas(self) -> None:
        """
        Check for captchas on the page and solve them.
        """
        await self._notify_observers(
            "on_check_and_solve_captchas", self.page, check_duplication=False
        )

    async def _notify_observers(
        self,
        method: ObservationTrigger,
        *args: Any,
        check_duplication: bool = True,
        **kwargs: Any,
    ) -> Any:
        """
        Notify all observers of an event. This will call the method on each observer that is subscribed. Note that
        the first observer is the stop pagination observer, so it will always be called separately so that we can stop
        pagination if needed.

        :param method: observation trigger
        :param args: arguments to pass to the method
        :param kwargs: keyword arguments to pass to the method
        :return: the result of the method call
        """
        duplicated = False
        if check_duplication:
            duplicated = getattr(self._deduper, method)(*args, **kwargs)
        if not duplicated:
            return await asyncio.gather(
                *[getattr(o, method)(*args, **kwargs) for o in self._observers]
            )

    @staticmethod
    async def run(
        scraper: AsyncScraperType,
        url: str | Path,
        schema: Schema | None = None,
        context: Optional[Context] = None,
        setup: Optional[SetupType] = None,
        harness: WebHarness = playwright_harness,
        evaluator: Optional[ExpressionEvaluator] = None,
        observer: Optional[OutputObserver | List[OutputObserver]] = None,
        goto_error_handler: Callable[
            [str, int, dict[str, str]], Awaitable[None]
        ] = default_error_callback,
        **harness_options: Unpack[HarnessOptions],
    ) -> "SDK":
        """
        Convenience method for running a scraper. This will launch a browser and
        invoke the scraper function.
        :param scraper: scraper to run
        :param url: starting url to run the scraper on
        :param schema: schema used to validate output correctness
        :param context: additional context to pass to the scrapers
        :param setup: setup function to run before the scraper
        :param harness: the harness to use for the browser
        :param evaluator: expression evaluator to use for the scraper
        :param observer: observer to use for the scraper
        :return none: everything should be saved to the database or file
        """
        domain = getattr(scraper, "domain", None)
        stage = getattr(scraper, "stage", None)
        observer = observer or getattr(scraper, "observer", None)
        context = context or {}

        harness_options.setdefault("headers", getattr(scraper, "extra_headers", None))  # type: ignore

        if isinstance(url, Path):
            url = f"file://{url.resolve()}"

        async with harness(**harness_options) as page_factory:
            page = await page_factory()
            sdk = SDK(
                page,
                domain=domain,
                stage=stage,
                observer=observer,
                scraper=scraper,
                context=context,
                schema=schema,
                evaluator=evaluator,
            )
            if setup:
                await setup(sdk)

            if not harness_options.get("disable_go_to_url", False):
                response = await page.goto(url)
                if response.status >= 400:
                    await goto_error_handler(url, response.status, response.headers)
            elif isinstance(page, SoupPage):
                page.url = url
            await scraper(sdk, url, context)

        return sdk

    async def get_content_type(self, url: str) -> str:
        async with aiohttp.ClientSession() as session:
            async with session.head(normalize_url(url, self.page.url)) as response:
                return response.headers.get("Content-Type", "")

    @staticmethod
    async def run_from_file(
        scraper: AsyncScraperType,
        schema: Schema,
        setup: Optional[SetupType] = None,
        **harness_options: Unpack[HarnessOptions],
    ) -> "SDK":
        """
        Convenience method for running a detail scraper from file. This will load
        the listing data from file and pass it to the scraper.

        :param scraper: the scraper to run (function)
        :param schema: schema used to validate output correctness
        :param headless: whether to run the browser headless
        :param cdp_endpoint: endpoint to connect to the browser (if using a remote browser)
        :param proxy: proxy to use for the browser
        :param setup: optional setup
        :return: None: the scraper should save data to the database or file
        """
        domain: str = getattr(scraper, "domain", "")
        stage: str = getattr(scraper, "stage", "")
        headers: dict[str, str] = getattr(scraper, "extra_headers", {})
        observer: Optional[OutputObserver] = getattr(scraper, "observer", None)

        if stage != "detail" and stage != "listing":
            raise ValueError("Only listing / detail scrapers can be run from file")

        tracker = FileDataTracker(domain, stage)

        prev = "listing" if stage == "detail" else "category"
        file_path = tracker.get_storage_filepath(prev)

        if not file_path.exists():
            raise ValueError(
                f"Could not find {file_path}."
                f" No listing data found for this domain. Run the listing scraper first."
            )

        listing_data = tracker.load_data(domain, prev)
        async with playwright_harness(**harness_options) as page_factory:
            page = await page_factory()
            for listing in listing_data:
                sdk = SDK(
                    page,
                    domain=domain,
                    stage=stage,  # type: ignore
                    observer=observer,
                    scraper=scraper,
                    schema=schema,
                )
                if setup:
                    await setup(sdk)

                if headers:
                    await page.set_extra_http_headers(headers)
                await page.goto(listing["url"])
                await scraper(
                    sdk,
                    listing["url"],
                    listing["context"],
                )

        return sdk

    @staticmethod
    def scraper(
        domain: str,
        stage: Stage,
        observer: Optional[OutputObserver | List[OutputObserver]] = None,
    ) -> Callable[[AsyncScraperType], AsyncScraperType]:
        """
        Decorator for scrapers. This will add the domain and stage to the function.
        All scrapers should be decorated with this decorator.
        :param domain: the url that the scraper is for (eg: https://example.org)
        :param stage: the stage of the scraper (eg: listing or detail)
        :param observer: the observer to use for the scraper
        :return: the decorated function
        """
        if not observer:
            observer = [LocalStorageObserver(domain, stage), LoggingObserver()]
        if not isinstance(observer, list):
            observer = [observer]

        def decorator(func: AsyncScraperType) -> AsyncScraperType:
            @wraps(func)
            async def wrapper(sdk: "SDK", url: URL, context: Context) -> None:
                return await func(sdk, url, context)

            wrapper.domain = domain  # type: ignore
            wrapper.stage = stage  # type: ignore
            wrapper.observer = observer  # type: ignore
            return wrapper

        return decorator

    @staticmethod
    def with_headers(
        headers: dict[str, str],
    ) -> Callable[[AsyncScraperType], AsyncScraperType]:
        """
        Decorator for scrapers. This will add the headers to the function.
        All scrapers should be decorated with this decorator.
        :param headers: the headers to use for the scraper
        :return: the decorated function
        """

        def decorator(func: AsyncScraperType) -> AsyncScraperType:
            @wraps(func)
            async def wrapper(sdk: "SDK", url: URL, context: Context) -> None:
                return await func(sdk, url, context)

            wrapper.extra_headers = headers  # type: ignore
            return wrapper

        return decorator


PAGE_PDF_FILENAME = "reworkd_page_pdf.pdf"



import pytest

from harambe.pagination import PaginatedList, DuplicateHandler, PageInfo


@pytest.fixture
def paginated_collection():
    return PaginatedList()


@pytest.fixture
def duplicate_handler():
    return DuplicateHandler()


async def test_pagination_with_no_data_succeeds(duplicate_handler):
    assert duplicate_handler.get_number_of_pages() == 1
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=1, total_rows=0, duplicated_rows=0
    )

    duplicate_handler.on_paginate("https://example.com/page3")

    assert duplicate_handler.get_number_of_pages() == 2
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=0, duplicated_rows=0
    )


async def test_stop_pagination_observer_duplicate_data_error(duplicate_handler):
    unduplicated = duplicate_handler.on_save_data({"foo": "bar"})
    duplicate_handler.on_paginate("https://example.com/page2")
    duplicated = duplicate_handler.on_save_data({"foo": "bar"})

    assert not unduplicated and duplicated
    assert duplicate_handler.get_number_of_pages() == 2
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=1, duplicated_rows=1
    )

    with pytest.raises(StopAsyncIteration):
        duplicate_handler.on_paginate("https://example.com/page3")


async def test_stop_pagination_observer_duplicate_url_error(duplicate_handler):
    unduplicated = duplicate_handler.on_queue_url(
        "https://example.com", {"foo": "bar"}, {}
    )
    duplicate_handler.on_paginate("https://example.com/page2")
    duplicated = duplicate_handler.on_queue_url(
        "https://example.com", {"foo": "bar"}, {}
    )

    assert not unduplicated and duplicated
    assert duplicate_handler.get_number_of_pages() == 2
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=1, duplicated_rows=1
    )

    with pytest.raises(StopAsyncIteration):
        duplicate_handler.on_paginate("https://example.com/page3")


async def test_stop_pagination_observer_duplicate_download_error(duplicate_handler):
    unduplicated = duplicate_handler.on_download(
        "https://example.com", "foo.txt", b"foo"
    )
    duplicate_handler.on_paginate("https://example.com/page2")
    duplicated = duplicate_handler.on_download("https://example.com", "foo.txt", b"foo")

    assert not unduplicated and duplicated
    assert duplicate_handler.get_number_of_pages() == 2
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=1, duplicated_rows=1
    )

    with pytest.raises(StopAsyncIteration):
        duplicate_handler.on_paginate("https://example.com/page3")


async def test_stop_pagination_observer_no_duplicate_data(duplicate_handler):
    unduplicated1 = duplicate_handler.on_save_data({"foo": "bar"})
    duplicate_handler.on_paginate("https://example.com/page2")
    unduplicated2 = duplicate_handler.on_save_data({"baz": "qux"})
    unduplicated3 = duplicate_handler.on_save_data(
        {
            "foo": [
                "bar",
                "baz",
            ]
        }
    )

    assert not unduplicated1 and not unduplicated2 and not unduplicated3
    assert duplicate_handler.get_number_of_pages() == 2
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=2, duplicated_rows=0
    )


async def test_ignore_underscore_attributes(duplicate_handler):
    unduplicated1 = duplicate_handler.on_save_data({"foo": "bar", "__url": "qux"})
    unduplicated2 = duplicate_handler.on_save_data({"qux": "bar", "__url": "qux"})
    duplicate_handler.on_paginate("https://example.com/page2")

    duplicated1 = duplicate_handler.on_save_data(
        {"foo": "bar", "__url": "bad boy asim"}
    )
    duplicated2 = duplicate_handler.on_save_data(
        {"qux": "bar", "__url": "bad boy adam"}
    )

    assert not unduplicated1 and not unduplicated2 and duplicated1 and duplicated2
    assert duplicate_handler.get_number_of_pages() == 2
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=2, duplicated_rows=2
    )

    with pytest.raises(StopAsyncIteration):
        duplicate_handler.on_paginate("https://example.com/page3")


async def test_duplicate_data_without_pagination(duplicate_handler):
    un_duplicated = duplicate_handler.on_save_data({"foo": "bar"})
    duplicated = duplicate_handler.on_save_data({"foo": "bar"})
    assert not un_duplicated and duplicated
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=1, total_rows=2, duplicated_rows=1
    )

    un_duplicated = duplicate_handler.on_queue_url(
        "https://example.com", {"foo": "bar"}, {}
    )
    duplicated = duplicate_handler.on_queue_url(
        "https://example.com", {"foo": "bar"}, {}
    )
    assert not un_duplicated and duplicated
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=1, total_rows=4, duplicated_rows=2
    )

    un_duplicated = duplicate_handler.on_download(
        "https://example.com", "foo.txt", b"foo"
    )
    duplicated = duplicate_handler.on_download("https://example.com", "foo.txt", b"foo")
    assert not un_duplicated and duplicated
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=1, total_rows=6, duplicated_rows=3
    )

    duplicate_handler.on_paginate("https://example.com/page2")
    assert duplicate_handler.get_number_of_pages() == 2

    duplicated = duplicate_handler.on_save_data({"foo": "bar"})
    assert duplicated
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=1, duplicated_rows=1
    )

    with pytest.raises(StopAsyncIteration):
        duplicate_handler.on_paginate("https://example.com/page3")


def test_append(paginated_collection):
    assert not len(paginated_collection)
    assert paginated_collection.should_continue()

    paginated_collection.append(1)
    assert len(paginated_collection) == 1
    assert paginated_collection.should_continue()

    paginated_collection.append(2)
    assert len(paginated_collection) == 2
    assert paginated_collection.should_continue()

    paginated_collection.append(1)
    assert len(paginated_collection) == 2
    assert not paginated_collection.should_continue()


def test_extend(paginated_collection):
    paginated_collection.extend(
        [{"foo": "bar"}, {"baz": "qux"}, {"foo": "bar", "baz": "qux"}]
    )
    assert len(paginated_collection) == 3
    assert paginated_collection.should_continue()

    paginated_collection.extend([{"adam": "eve"}, {"adam": {"eve": "adam"}}])
    assert len(paginated_collection) == 5
    assert paginated_collection.should_continue()

    paginated_collection.extend([{"foo": "bar", "baz": "qux"}, {"eve": "adam"}])
    assert len(paginated_collection) == 6
    assert not paginated_collection.should_continue()


def test_get_number_of_pages(duplicate_handler):
    duplicate_handler.on_save_data({"foo": "bar"})
    assert duplicate_handler.get_number_of_pages() == 1

    duplicate_handler.on_paginate("https://example.com/page2")
    assert duplicate_handler.get_number_of_pages() == 2

    duplicate_handler.on_save_data({"baz": "qux"})
    assert duplicate_handler.get_number_of_pages() == 2

    duplicate_handler.on_paginate("https://example.com/page3")
    assert duplicate_handler.get_number_of_pages() == 3


def test_get_current_page_info(duplicate_handler):
    duplicate_handler.on_save_data({"foo": "bar"})
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=1, total_rows=1, duplicated_rows=0
    )

    duplicate_handler.on_save_data({"foo": "bar"})
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=1, total_rows=2, duplicated_rows=1
    )

    duplicate_handler.on_paginate("https://example.com/page2")
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=0, duplicated_rows=0
    )

    duplicate_handler.on_save_data({"baz": "qux"})
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=1, duplicated_rows=0
    )

    duplicate_handler.on_save_data({"foo": "bar"})
    assert duplicate_handler.get_current_page_info() == PageInfo(
        page=2, total_rows=2, duplicated_rows=1
    )



import asyncio

from harambe.user_agent import compute_user_agent, random_user_agent


async def test_user_agent_with_string():
    result = await compute_user_agent("my-user-agent")
    assert result == "my-user-agent"


async def test_user_agent_with_sync_callable():
    def sync_callable() -> str:
        return "sync-user-agent"

    result = await compute_user_agent(sync_callable)
    assert result == "sync-user-agent"

    res = await compute_user_agent(random_user_agent)
    assert res
    assert isinstance(res, str)


async def test_user_agent_with_async_callable():
    async def async_callable() -> str:
        await asyncio.sleep(0.1)
        return "async-user-agent"

    result = await compute_user_agent(async_callable)
    assert result == "async-user-agent"


def test_default_user_agent():
    res = random_user_agent()
    assert res
    assert isinstance(res, str)



import pytest

from harambe.contrib import WebHarness, playwright_harness, soup_harness


@pytest.fixture(params=[playwright_harness, soup_harness])
def web_harness(request) -> WebHarness:
    return request.param


async def test_with_cookies(web_harness):
    async with web_harness(
        cookies=[
            {"url": "https://example.com", "name": "foo", "value": "bar"},
            {
                "url": "https://example.com",
                "name": "baz",
                "value": "qux",
            },
        ]
    ) as page_factory:
        assert page_factory


async def test_with_user_agent():
    async with playwright_harness(user_agent=lambda: "my-user-agent") as page_factory:
        page = await page_factory()
        user_agent = await page.evaluate("navigator.userAgent")
        assert user_agent == "my-user-agent"


async def test_default_url(web_harness):
    async with web_harness() as page_factory:
        page = await page_factory()
        assert page.url == "about:blank"


async def test_with_two_connections():
    async with (
        playwright_harness(launch_args=["--remote-debugging-port=9222"]) as p1,
        playwright_harness(cdp_endpoint="http://localhost:9222") as p2,
    ):
        page_1 = await p1()
        await page_1.goto("https://example.com/")

        page_2 = await p2()
        assert page_2.url == "about:blank"
        assert page_2.context.browser.contexts[0].pages[0].url == "https://example.com/"

        await (
            page_2.context.browser.contexts[0].pages[0].goto("https://www.reworkd.ai/")
        )
        assert page_1.url == "https://www.reworkd.ai/"



from harambe.types import Cookie
from datetime import datetime, timezone, timedelta


def fix_cookie(cookie: Cookie) -> Cookie:
    """
    Fix the cookie expiry by setting it to one day ahead only if it's expiring within a day.

    :param cookie: The cookie to modify.
    :return: The modified cookie.
    """
    current_time = datetime.now(tz=timezone.utc)
    if "expires" in cookie:
        expiry_time = datetime.fromtimestamp(cookie["expires"], tz=timezone.utc)
        if (expiry_time - current_time) < timedelta(days=1):
            cookie["expires"] = (current_time + timedelta(days=1)).timestamp()
    else:
        cookie["expires"] = (current_time + timedelta(days=1)).timestamp()

    return cookie



from pathlib import Path
from typing import cast

import pytest
from aiohttp import web
from bs4 import BeautifulSoup

from harambe import SDK
from harambe.contrib import playwright_harness, soup_harness
from harambe.observer import InMemoryObserver
from harambe.types import BrowserType
from harambe_core.errors import GotoError


@pytest.fixture(scope="module")
def mock_html_folder():
    path = Path(__file__).parent / "mock_html"
    assert path.exists() and path.is_dir() and len(list(path.iterdir())) > 0
    return path


@pytest.fixture
async def server(mock_html_folder):
    async def handle(request):
        if request.path == "/403":
            return web.Response(text="Forbidden", status=403)

        file_path = mock_html_folder / f"{request.path.strip('/')}.html"
        if not file_path.exists():
            return web.FileResponse(mock_html_folder / "404.html")

        return web.FileResponse(file_path)

    app = web.Application()
    app.router.add_route("*", "/{tail:.*}", handle)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, "127.0.0.1", 8081)
    await site.start()

    yield site.name.strip("/")

    await runner.cleanup()


@pytest.fixture
def observer():
    return InMemoryObserver()


@pytest.mark.parametrize("browser_type", ["chromium", "firefox", "webkit"])
@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_save_data(server, observer, harness, browser_type):
    url = f"{server}/table"

    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page

        await page.wait_for_load_state()
        for row in await page.query_selector_all("tbody > tr"):
            fruit, price = await row.query_selector_all("td")

            await sdk.save_data(
                {"fruit": await fruit.inner_text(), "price": await price.inner_text()}
            )

    browser_type = cast(BrowserType, browser_type)
    await SDK.run(
        scraper=scraper,
        url=url,
        schema={},
        headless=True,
        harness=harness,
        browser_type=browser_type,
        observer=observer,
    )

    assert len(observer.data) == 3

    assert observer.data[0]["fruit"] == "Apple"
    assert observer.data[0]["price"] == "1.00"
    assert observer.data[0]["__url"] == url

    assert not observer.urls
    assert not observer.files


async def test_enqueue_data(server, observer):
    async def scraper(sdk: SDK, *args, **kwargs):
        await sdk.enqueue("?page=1")
        await sdk.enqueue("/terms", "https://reworkd.ai")

    await SDK.run(
        scraper=scraper, url=server, schema={}, headless=True, observer=observer
    )

    assert not observer.data
    assert len(observer.urls) == 3
    assert observer.urls[0][0] == f"{server}/?page=1"
    assert observer.urls[1][0] == f"{server}/terms"
    assert observer.urls[2][0] == "https://reworkd.ai"

    url = f"{server}/"
    assert observer.urls[0][1] == {"__url": url}
    assert observer.urls[1][1] == {"__url": url}
    assert observer.urls[2][1] == {"__url": url}


async def test_enqueue_data_with_context(server, observer):
    async def scraper(sdk: SDK, *args, **kwargs):
        await sdk.enqueue("/adam/?page=55", context={"last": "Watkins"})

    await SDK.run(
        scraper=scraper, url=server, schema={}, headless=True, observer=observer
    )

    assert not observer.data
    assert len(observer.urls) == 1
    assert observer.urls[0][0] == f"{server}/adam/?page=55"
    assert observer.urls[0][1] == {"__url": f"{server}/", "last": "Watkins"}


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_enqueue_no_goto_url(observer, harness):
    async def scraper(sdk: SDK, *args, **kwargs):
        await sdk.enqueue("https://reworkd.ai", context={"last": "Watkins"})

    await SDK.run(
        scraper=scraper,
        url="https://example.com",
        schema={},
        headless=True,
        disable_go_to_url=True,
        harness=harness,
        observer=observer,
    )
    assert len(observer.urls) == 1
    assert observer.urls[0][0] == "https://reworkd.ai"


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_enqueue_coro(server, observer, harness):
    @SDK.scraper("test", "detail", observer=observer)
    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        await sdk.enqueue((await page.query_selector("a")).get_attribute("href"))

    await SDK.run(
        scraper=scraper,
        url=f"{server}/table",
        schema={},
        headless=True,
        harness=harness,
    )
    assert len(observer.urls) == 1
    assert observer.urls[0][0] == f"{server}/other"


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_paginate(server, observer, harness):
    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        await sdk.save_data({"content": await page.content()})

        async def pager():
            link = await page.query_selector("a")
            return await link.get_attribute("href")

        await sdk.paginate(pager)

    await SDK.run(
        scraper=scraper,
        url=f"{server}/table",
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert len(observer.data) == 2
    assert observer.data[0]["content"]
    assert observer.data[0]["__url"] == f"{server}/table"

    assert observer.data[1]["content"]
    assert observer.data[1]["__url"] == f"{server}/other"


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_narcotics(server, observer, harness):
    async def scraper(sdk: SDK, *args, **kwargs) -> None:
        page = sdk.page
        await page.wait_for_selector("div#ds-content")

        document_url_elements = await page.query_selector_all("div.file-wrapper")

        data = []
        for doc in document_url_elements:
            title_element = await doc.query_selector("div.file-metadata span[title]")
            title = (await title_element.get_attribute("title")).replace(".pdf", "")

            document_url_element = await doc.query_selector("a.image-link")
            document_url = await document_url_element.get_attribute("href")

            data.append({"title": title, "document_url": document_url})

        for item in data:
            await sdk.save_data(item)

    await SDK.run(
        scraper=scraper,
        url=f"{server}/narcotics",
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert len(observer.data) == 1
    assert observer.data[0] == {
        "title": "NARCOTICS CONTROL COMMISSION ACT, 2020 (ACT 1019)",
        "document_url": "/bitstream/handle/123456789/1921/meth.pdf?sequence=1&isAllowed=y",
        "__url": f"{server}/narcotics",
    }


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_regulations(server, observer, harness):
    async def scraper(sdk: SDK, *args, **kwargs) -> None:
        page = sdk.page
        await page.wait_for_selector("table.table.mb-0.table-hover.table-striped")

        links = await page.query_selector_all(
            "a[title='Go to the Regulations page'] + ul a"
        )
        for link in links:
            href = await link.get_attribute("href")
            await sdk.enqueue(href)

    await SDK.run(
        scraper=scraper,
        url=f"{server}/regulations",
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert not observer.data
    assert len(observer.urls) == 3
    assert observer.urls[0][0] == f"{server}/regulations/act/"
    assert observer.urls[1][0] == f"{server}/regulations/regulations/"
    assert observer.urls[2][0] == f"{server}/regulations/guidelines/"
    assert observer.urls[0][1] == {"__url": f"{server}/regulations"}
    assert observer.urls[1][1] == {"__url": f"{server}/regulations"}
    assert observer.urls[2][1] == {"__url": f"{server}/regulations"}


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_text_content(server, observer, harness):
    url = f"{server}/table"

    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        content = await page.text_content("table")
        await sdk.save_data({"page_content": content})

        table = await page.query_selector("table")
        await sdk.save_data({"table_content": await table.text_content()})

    await SDK.run(
        scraper=scraper,
        url=url,
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )
    assert len(observer.data) == 2

    assert observer.data[0]["page_content"] == observer.data[1]["table_content"]
    for text in ["Apple", "Orange", "Banana"]:
        assert text in observer.data[0]["page_content"], (
            f"{text} not in {observer.data[0]['page_content']}"
        )
        assert text in observer.data[1]["table_content"], (
            f"{text} not in {observer.data[1]['table_content']}"
        )


@pytest.mark.parametrize("harness", [soup_harness])
async def test_text_content_when_selector_does_not_exist(server, observer, harness):
    url = f"{server}/table"

    @SDK.scraper("test", "detail", observer=observer)
    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        await page.set_default_timeout(1)

        content = await page.text_content("table.non-existent")
        await sdk.save_data({"page_content": content})

    await SDK.run(scraper=scraper, url=url, schema={}, headless=True, harness=harness)
    assert len(observer.data) == 1
    assert observer.data[0]["page_content"] is None


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_page_goto_with_options(server, harness):
    url = f"{server}/table"

    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        await page.goto(
            "https://example.com", timeout=1000, referer="https://google.com"
        )

    await SDK.run(scraper=scraper, url=url, schema={}, headless=True, harness=harness)


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_currency_validator(server, harness):
    async def scraper(sdk: SDK, *args, **kwargs):
        await sdk.save_data({"price": "$1,9999.00"})

    await SDK.run(
        scraper=scraper,
        url=f"{server}/table",
        schema={"price": {"type": "currency"}},
        headless=True,
        harness=harness,
    )


@pytest.mark.parametrize("harness", [soup_harness])
async def test_disable_go_to_url_bug(server, harness):
    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        await sdk.save_data({"url": page.url})

    await SDK.run(
        scraper=scraper,
        disable_go_to_url=True,
        url="https://www.bcp.gov.gh/acc/registry/docs/Ghana%20Export%20Promotion%20Authority%20Act,%201969%20(NLCD%20396).pdf",
        schema={"url": {"type": "url"}},
        headless=True,
        harness=harness,
    )


@pytest.mark.parametrize("harness", [playwright_harness])
async def test_save_local_storage(server, observer, harness):
    local_storage_entry = {
        "domain": "asim-shrestha.com",
        "path": "/",
        "key": "test_key",
        "value": "test",
    }

    @SDK.scraper(local_storage_entry["domain"], "detail")
    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        # Save test local storage key value pair
        await page.evaluate(
            f"localStorage.setItem('{local_storage_entry['key']}', '{local_storage_entry['value']}')"
        )

        await sdk.save_local_storage()

    await SDK.run(
        scraper=scraper,
        url=f"https://{local_storage_entry['domain']}/",
        headless=True,
        harness=harness,
        schema={},
        observer=observer,
    )

    assert len(observer.local_storage) == 1
    assert observer.local_storage == [local_storage_entry]


@pytest.mark.parametrize("harness", [playwright_harness])
@pytest.mark.parametrize(
    "test_value,expected_value",
    [
        # String value
        ("test_string", "test_string"),
        # Number value
        (42, "42"),
        # List value
        (["item1", "item2", 3], '["item1", "item2", 3]'),
        # Dict value
        ({"key1": "value1", "key2": 2}, '{"key1": "value1", "key2": 2}'),
        # Nested structure
        (
            {"list": [1, 2, {"nested": "value"}]},
            '{"list": [1, 2, {"nested": "value"}]}',
        ),
    ],
)
async def test_load_local_storage(
    server, observer, harness, test_value, expected_value
):
    local_storage_entry_1 = {
        "domain": "asim-shrestha.com",
        "path": "/",
        "key": "test_key",
        "value": test_value,
    }

    local_storage_entry_2 = {
        "domain": "asim-shrestha.com",
        "path": "/",
        "key": "another_key",
        "value": test_value,
    }

    async def scraper(sdk: SDK, *args, **kwargs):
        page = sdk.page
        page_local_storage = await page.evaluate("localStorage")
        await sdk.save_data({"local_storage": page_local_storage})

    await SDK.run(
        scraper=scraper,
        url=f"https://{local_storage_entry_1['domain']}/",
        headless=True,
        harness=harness,
        schema={},
        local_storage=[local_storage_entry_1, local_storage_entry_2],
        observer=observer,
    )

    assert len(observer.data) == 1
    assert observer.data[0]["local_storage"] == {
        local_storage_entry_1["key"]: expected_value,
        local_storage_entry_2["key"]: expected_value,
    }


@pytest.mark.parametrize("harness", [playwright_harness])
async def test_reset_local_storage(server, observer, harness):
    local_storage_entry = {
        "domain": "asim-shrestha.com",
        "path": "/",
        "key": "test_key",
        "value": "test_value",
    }

    async def scraper(sdk: SDK, current_url: str, *args, **kwargs):
        page = sdk.page
        await page.evaluate("localStorage.clear();")
        await page.goto(current_url)
        page_local_storage = await page.evaluate("localStorage")
        await sdk.save_data({"local_storage": page_local_storage})

    await SDK.run(
        scraper=scraper,
        url=f"https://{local_storage_entry['domain']}/",
        headless=True,
        harness=harness,
        schema={},
        local_storage=[local_storage_entry, local_storage_entry],
        observer=observer,
    )

    assert len(observer.data) == 1
    assert observer.data[0]["local_storage"] == {}


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_capture_html_with_different_options(server, observer, harness):
    url = f"{server}/table"

    replaced_element = '<div id="reworkd">Replaced Text</div>'

    async def scraper(sdk: SDK, *args, **kwargs):
        full_html_metadata = await sdk.capture_html()
        await sdk.save_data(full_html_metadata)

        table_html_metadata = await sdk.capture_html("table", ["thead"])
        await sdk.save_data(table_html_metadata)

        table_head_with_replaced_text_html_metadata = await sdk.capture_html(
            "table",
            soup_transform=lambda soup: soup.find("thead").replace_with(
                BeautifulSoup(replaced_element, "html.parser")
            ),
        )
        await sdk.save_data(table_head_with_replaced_text_html_metadata)

    await SDK.run(
        scraper=scraper,
        url=url,
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert len(observer.data) == 3

    # Verify full document capture
    doc_data = observer.data[0]
    assert doc_data["html"].startswith("<!DOCTYPE html>")
    assert doc_data["html"].count("<!DOCTYPE html>") == 1
    assert "<table" in doc_data["html"]
    assert "<tbody" in doc_data["html"]
    assert replaced_element not in doc_data["html"]
    assert "Apple" in doc_data["text"]

    # Verify download fields all available
    assert doc_data["url"]
    assert doc_data["filename"]

    # Verify table capture with exclusion
    table_data = observer.data[1]
    assert table_data["html"].startswith("<!DOCTYPE html>")
    assert table_data["html"].count("<!DOCTYPE html>") == 1
    assert "<tbody" in doc_data["html"]
    assert "<thead" not in table_data["html"]
    assert "Price" not in table_data["text"]
    assert "Apple" in table_data["text"]

    replaced_head_data = observer.data[2]
    assert replaced_head_data["html"].count("<!DOCTYPE html>") == 1
    assert replaced_head_data["html"].startswith("<!DOCTYPE html>")
    assert "<thead" not in replaced_head_data["html"]
    assert replaced_element in replaced_head_data["html"]
    assert "Replaced Text" in replaced_head_data["text"]


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_capture_html_conversion_types(server, observer, harness):
    url = f"{server}/heading"

    async def scraper(sdk: SDK, *args, **kwargs):
        markdown_html_metadata = await sdk.capture_html()
        await sdk.save_data({"text": markdown_html_metadata["text"]})

        text_html_metadata = await sdk.capture_html(html_converter_type="text")
        await sdk.save_data({"text": text_html_metadata["text"]})

    await SDK.run(
        scraper=scraper,
        url=url,
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert len(observer.data) == 2
    # Markdown syntax is used
    assert observer.data[0]["text"].strip() == "### Heading"

    # Text doesn't include markdown syntax
    assert observer.data[1]["text"].strip() == "Heading"


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_capture_html_table(server, observer, harness):
    url = f"{server}/table"

    async def scraper(sdk: SDK, *args, **kwargs):
        text_html_metadata = await sdk.capture_html(html_converter_type="text")
        await sdk.save_data({"text": text_html_metadata["text"]})

    await SDK.run(
        scraper=scraper,
        url=url,
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert len(observer.data) == 1
    assert observer.data[0]["text"].strip() == (
        "Food Prices\n\n"
        "Shown below are the prices of some fruits:\n\n"
        "| Food | Price |\n"
        "| --- | --- |\n"
        "| Apple | 1.00 |\n"
        "| Banana | 0.50 |\n"
        "| Orange [10] | 1.25 |\n\n"
        "Other Page"
    )


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_capture_html_element_not_found(server, observer, harness):
    url = f"{server}/table"

    async def scraper(sdk: SDK, *args, **kwargs):
        with pytest.raises(ValueError):
            await sdk.capture_html("#missing .selector .lies .adam-watkins")

    await SDK.run(
        scraper=scraper,
        url=url,
        schema={},
        headless=True,
        harness=harness,
        observer=observer,
    )

    assert len(observer.data) == 0


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_with_locators(server, observer, harness):
    url = f"{server}/solicitation"

    async def scrape(sdk: SDK, url, context) -> None:
        from harambe.utils import PlaywrightUtils as Pu

        page = sdk.page
        await page.wait_for_selector("span#ctl00_MainBody_lblBidNo")

        solicitation_id = await Pu.get_text(page, "span#ctl00_MainBody_lblBidNo")
        title = await Pu.get_text(page, "span#ctl00_MainBody_lblBidTitle")
        description = await Pu.get_text(page, "span#ctl00_MainBody_lblDesc")

        attachments = []
        attachment_titles = await Pu.get_texts(
            page, "table#ctl00_MainBody_dgFileList a"
        )
        attachment_urls = await Pu.get_links(page, "table#ctl00_MainBody_dgFileList a")
        for attachment_title, attachment_url in zip(attachment_titles, attachment_urls):
            attachments.append({"title": attachment_title, "url": attachment_url})

        await sdk.save_data(
            {
                "solicitation_id": solicitation_id,
                "title": title,
                "description": description,
                "attachments": attachments,
                "status": context.get("status"),
            }
        )

    await SDK.run(
        scrape,
        url,
        schema={},
        harness=harness,
        context={"status": "Open"},
        observer=observer,
    )
    assert len(observer.data) == 1
    assert observer.data[0]["solicitation_id"] == "6100062375"
    assert observer.data[0]["title"] == "23SW SGL 111 Conn Road"
    assert (
        observer.data[0]["description"]
        == "The State of Pennsylvania is seeking proposals for IT services"
    )
    assert observer.data[0]["status"] == "Open"
    assert len(observer.data[0]["attachments"]) == 4


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
async def test_403_status_on_goto_with_default_callback(server, observer, harness):
    url = f"{server}/403"

    async def scrape(sdk: SDK, current_url, context) -> None:
        await sdk.save_data({"key": "this shouldn't be saved if GotoError is raised"})

    with pytest.raises(GotoError):
        await SDK.run(
            scraper=scrape,
            url=url,
            harness=harness,
            schema={},
            context={"status": "Open"},
            observer=observer,
        )
    assert len(observer.data) == 0


@pytest.mark.parametrize("harness", [playwright_harness, soup_harness])
@pytest.mark.parametrize("goto_error_cb", ["custom"])
async def test_403_status_on_goto_with_custom_callback(
    server, observer, harness, goto_error_cb
):
    url = f"{server}/403"

    async def scrape(sdk: SDK, current_url, context) -> None:
        await sdk.save_data({"key": "this shouldn't be saved if GotoError is raised"})

    async def custom_error_handler(url, status_code, *args):
        print(f"Handled {status_code} for {url} gracefully.")

    error_callback = custom_error_handler
    await SDK.run(
        scraper=scrape,
        url=url,
        harness=harness,
        schema={},
        context={"status": "Open"},
        observer=observer,
        goto_error_handler=error_callback,
    )

    # Ensure data is saved when error is handled (either with custom or no callback)
    assert len(observer.data) == 1
    assert observer.data[0]["key"] == "this shouldn't be saved if GotoError is raised"
    assert observer.data[0]["__url"] == url



from urllib.parse import urlparse

from playwright.async_api import ProxySettings


def proxy_from_url(url: str) -> ProxySettings:
    parsed = urlparse(url, allow_fragments=False)

    if not parsed.hostname:
        parsed = urlparse(f"http://{url}", allow_fragments=False)

    if not all(
        [
            parsed.hostname,
            parsed.username,
            parsed.password,
        ]
    ):
        raise ValueError(f"Invalid proxy URL: {url}")

    proxy: ProxySettings = {
        "server": parsed.hostname,  # type: ignore
        "username": parsed.username,
        "password": parsed.password,
    }

    if parsed.port:
        proxy["server"] += f":{parsed.port}"

    return proxy



from unittest.mock import AsyncMock, call

import pytest
from playwright.async_api import Page

from harambe.core import SDK, URL, AsyncScraperType, Context
from harambe.observer import OutputObserver
from harambe_core.errors import SchemaValidationError
from harambe_core import Schema


@pytest.fixture
def scraper() -> AsyncScraperType:
    async def dummy_scraper(sdk: "SDK", url: URL, context: Context) -> None:
        await sdk.page.goto(url)
        await sdk.enqueue("/test", context={"foo": "bar"})
        await sdk.save_data({"baz": "qux"})

    return dummy_scraper


def test_sdk_init_assigns_correct_values():
    page = AsyncMock(spec=Page)
    run_id = "test_run_id"
    observer = AsyncMock(spec=OutputObserver)

    sdk = SDK(page, run_id, "test_domain", "listing", observer)

    assert sdk.page == page
    assert sdk._id == run_id
    assert sdk._domain == "test_domain"
    assert sdk._stage == "listing"
    assert sdk._observers == [observer]


async def test_sdk_save_data_calls_on_save_data_for_each_observer():
    page = AsyncMock(spec=Page)
    observer = AsyncMock(spec=OutputObserver)
    sdk = SDK(page, observer=observer)
    data = [{"foo": "bar"}, {"baz": "qux"}]

    await sdk.save_data(*data)
    assert observer.on_save_data.call_count == len(data)


async def test_sdk_enqueue_calls_on_enqueue_url_for_each_observer():
    page = AsyncMock(spec=Page)
    page.url = "https://example.net"
    observer = AsyncMock(spec=OutputObserver)
    sdk = SDK(page, observer=observer)
    urls = ["https://example.org", "https://example.com"]
    context = {"foo": "bar"}
    options = {"zoo": "zar"}

    await sdk.enqueue(*urls, context=context, options=options)

    assert observer.on_queue_url.call_count == len(urls)
    observer.on_queue_url.assert_has_awaits(
        [call(urls[0], context, options), call(urls[1], context, options)],
        any_order=False,
    )

    assert observer.on_queue_url.call_count == len(urls)
    observer.on_save_data.assert_not_awaited()


# noinspection PyUnresolvedReferences
def test_scraper_decorator_adds_domain_and_stage_to_function(scraper):
    decorated_scraper = SDK.scraper("https://example.org", "listing")(scraper)

    assert decorated_scraper.domain == "https://example.org"
    assert decorated_scraper.stage == "listing"


# noinspection PyUnresolvedReferences
def test_scraper_decorator_adds_observers_to_function(scraper):
    domain = "https://example.org"

    decorated_scraper = SDK.scraper(domain, "detail")(scraper)
    assert len(decorated_scraper.observer) == 2


async def test_scraper_decorator_preserves_functionality_of_decorated_function(scraper):
    decorated_scraper = SDK.scraper("https://example.org", "listing")(scraper)

    sdk = AsyncMock(spec=SDK)
    sdk.page = AsyncMock()

    url = "https://example.org"
    context = {}

    await decorated_scraper(sdk, url, context)

    sdk.page.goto.assert_awaited_once_with(url)
    sdk.enqueue.assert_awaited_once_with("/test", context={"foo": "bar"})
    sdk.save_data.assert_awaited_once_with({"baz": "qux"})


async def test_sdk_save_data_saves_valid_data():
    page = AsyncMock(spec=Page)
    page.url = "https://example.net"
    observer = AsyncMock(spec=OutputObserver)
    schema: Schema = {
        "name": {"type": "string", "description": "The name of the person"},
        "age": {
            "type": "integer",
            "description": "Person's age",
        },
        "url": {
            "type": "url",
            "description": "Person's website",
        },
    }
    sdk = SDK(page, observer=observer, schema=schema)
    data = [
        {"name": "Joe Doe", "age": None, "url": "https://example.com"},
        {"name": "Jane Doe", "age": 33, "url": None},
    ]

    await sdk.save_data(*data)
    assert observer.on_save_data.call_count == 2


async def test_sdk_save_data_does_not_save_invalid_data():
    page = AsyncMock(spec=Page)
    observer = AsyncMock(spec=OutputObserver)
    schema: Schema = {"foo": {"type": "string", "description": "Something something"}}
    sdk = SDK(page, observer=observer, schema=schema)
    data = [{"baz": "456"}]

    with pytest.raises(SchemaValidationError):
        await sdk.save_data(*data)



from playwright.async_api import ElementHandle, Page

from harambe.contrib.types import AbstractElementHandle, AbstractPage


class PlaywrightElementHandle(ElementHandle, AbstractElementHandle):  # type: ignore
    pass


class PlaywrightPage(Page, AbstractPage[PlaywrightElementHandle]):  # type: ignore
    pass



import abc

# noinspection PyUnresolvedReferences,PyProtectedMember
from contextlib import _AsyncGeneratorContextManager
from typing import Any, Awaitable, Callable, Generic, Optional, TypeVar, Protocol

T = TypeVar("T", bound="AbstractElementHandle")
WebHarness = Callable[
    ..., _AsyncGeneratorContextManager[Callable[[], Awaitable["AbstractPage[Any]"]]]
]


class AbstractElementHandle(abc.ABC):
    @abc.abstractmethod
    async def inner_text(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def text_content(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def get_attribute(self, name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def click(self) -> None:
        raise NotImplementedError()


class Selectable(Generic[T], abc.ABC):
    @abc.abstractmethod
    async def query_selector_all(self, selector: str) -> list[T]:
        raise NotImplementedError()

    @abc.abstractmethod
    async def query_selector(self, selector: str) -> T | None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def wait_for_selector(self, selector: str, **kwargs: Any) -> None:
        raise NotImplementedError()


class ResponseWithStatus(Protocol):
    """Protocol for goto responses across all harnesses. Use minimal attributes required for current use cases."""

    status: int
    headers: dict[str, str]


class AbstractPage(Selectable[T], abc.ABC):
    @property
    @abc.abstractmethod
    def url(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def goto(self, url: str, **kwargs: Any) -> ResponseWithStatus:
        raise NotImplementedError()

    @abc.abstractmethod
    async def wait_for_timeout(self, timeout: int) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def content(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def text_content(self, selector: str, **kwargs: Any) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    async def set_extra_http_headers(self, headers: dict[str, str]) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def set_default_timeout(self, timeout: float) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def locator(self, selector: str) -> T | None:
        raise NotImplementedError()



import hashlib
import json
import zipfile
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from curl_cffi import requests

from harambe.utils import swallow_exceptions


def write_json_lines(file_path: Path, data: list[Any]) -> None:
    with file_path.open("w") as f:
        for item in data:
            f.write(json.dumps(item) + "\n")


class Tracer:
    def __init__(self) -> None:
        self._is_tracing = False
        self._network_traces: list[Any] = []
        self._resources: list[tuple[str, bytes]] = []

    def start(self) -> None:
        self._is_tracing = True

    def stop(self, *, path: str | Path) -> None:
        self._is_tracing = False
        self._create_zip(path)

    def _create_zip(self, path: str | Path) -> None:
        with TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)
            resources_path = temp_path / "resources"
            resources_path.mkdir(parents=True, exist_ok=True)

            # Create dummy trace files for demonstration
            trace_network_file = temp_path / "trace.network"
            trace_trace_file = temp_path / "trace.trace"
            trace_stacks_file = temp_path / "trace.stacks"

            write_json_lines(trace_network_file, self._network_traces)
            trace_trace_file.touch()
            trace_stacks_file.touch()

            # Write resources
            for file_name, content in self._resources:
                with (resources_path / file_name).open("wb") as f:
                    f.write(content)

            # Create the zip file
            with zipfile.ZipFile(path, "w") as zipf:
                for file_path in temp_path.rglob("*"):
                    zipf.write(file_path, file_path.relative_to(temp_path))

    @swallow_exceptions
    def log_request(self, response: requests.Response) -> None:
        if not self._is_tracing or not response.request:
            return

        started_date_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
        sha1 = hashlib.sha1(response.content).hexdigest()
        file_name = f"{sha1}.html"  # todo FILE types

        self._resources.append((file_name, response.content))
        self._network_traces.append(
            {
                "type": "resource-snapshot",
                "snapshot": {
                    "_frameref": "frame@7e64af4e3ffc9880713f6353fe66e81d",
                    "_monotonicTime": 0,  # TODO: calculate this
                    "startedDateTime": started_date_time,
                    "time": response.elapsed,
                    "request": {
                        "method": response.request.method,
                        "url": response.url,
                        "httpVersion": "HTTP/1.1",
                        "cookies": response.request.headers.get("Cookie", ""),
                        "headers": [
                            {"name": k, "value": v}
                            for k, v in response.request.headers.items()
                        ],
                        "queryString": [],
                        "headersSize": sum(
                            len(k) + len(v) for k, v in response.request.headers.items()
                        ),
                        "bodySize": 0,
                    },
                    "response": {
                        "status": response.status_code,
                        "statusText": response.reason,
                        "httpVersion": "HTTP/1.1",
                        "cookies": response.cookies.get_dict(),
                        "headers": [
                            {"name": k, "value": v} for k, v in response.headers.items()
                        ],
                        "content": {
                            "size": len(response.content),
                            "mimeType": response.headers.get("Content-Type", ""),
                            "compression": 0,
                            "_sha1": file_name,
                        },
                        "headersSize": sum(
                            len(k) + len(v) for k, v in response.headers.items()
                        ),
                        "bodySize": len(response.content),
                        "redirectURL": response.headers.get("Location", ""),
                        "_transferSize": len(response.content)
                        + sum(len(k) + len(v) for k, v in response.headers.items()),
                    },
                    "cache": {},
                    "timings": {
                        "dns": 0,  # Detailed timing info requires more complex instrumentation
                        "connect": 0,
                        "ssl": 0,
                        "send": 0,
                        "wait": response.elapsed,
                        "receive": 0,
                    },
                    "pageref": "page@cdd43826a397303886e152d3ad9548b3",
                    "serverIPAddress": response.primary_ip,
                    "_serverPort": 8081,
                    "_securityDetails": {},
                },
            }
        )



from typing import (
    Any,
    Awaitable,
    Callable,
    Literal,
    Optional,
    Sequence,
    TypedDict,
)

from playwright.async_api import ViewportSize

Enum = str
URL = str
ScrapeResult = dict[str, Any]
Context = dict[str, Any]
Options = dict[str, Any]
Stage = Literal["category", "listing", "detail"]
AsyncScraperType = Callable[["SDK", URL, Context], Awaitable[None]]  # type: ignore # noqa: F821
SetupType = Callable[["SDK"], Awaitable[None]]  # type: ignore # noqa: F821
Callback = Callable[..., Awaitable[None]]
BrowserType = Literal["chromium", "firefox", "webkit"]


class SetCookieParam(TypedDict, total=False):
    name: str
    value: str
    url: Optional[str]
    domain: Optional[str]
    path: Optional[str]
    expires: Optional[float]
    httpOnly: Optional[bool]
    secure: Optional[bool]
    sameSite: Optional[Literal["Lax", "None", "Strict"]]


class Cookie(TypedDict):
    name: str
    value: str
    domain: str
    path: str
    expires: int | float
    size: int
    httpOnly: bool
    secure: bool
    session: bool
    sameSite: str


class LocalStorage(TypedDict):
    domain: str
    path: str | None
    key: str
    value: str


class HarnessOptions(TypedDict, total=False):
    browser_type: BrowserType
    headless: bool
    stealth: bool
    default_timeout: int
    # TODO: Replace with UserAgentFactory
    user_agent: str | Callable[[], str | Awaitable[str]]
    proxy: Optional[str]
    cdp_endpoint: Optional[str]
    cookies: Sequence[SetCookieParam]
    local_storage: Sequence[LocalStorage]
    headers: Optional[dict[str, str]]
    viewport: Optional[ViewportSize]
    abort_unnecessary_requests: bool
    disable_go_to_url: bool
    on_start: Optional[Callback]
    on_end: Optional[Callback]
    extensions: Sequence[str]



import json
from collections import defaultdict
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Awaitable, Callable, Optional, Sequence, cast

from playwright.async_api import BrowserContext, ViewportSize, async_playwright, Page
from playwright_stealth import stealth_async

from harambe.contrib.playwright.impl import PlaywrightPage
from harambe.handlers import UnnecessaryResourceHandler
from harambe.proxy import proxy_from_url
from harambe.types import SetCookieParam, BrowserType, LocalStorage
from harambe.user_agent import random_user_agent, compute_user_agent, UserAgentFactory

Callback = Callable[[BrowserContext], Awaitable[None]]
PageCallback = Callable[[Page], Awaitable[None]]
PageFactory = Callable[..., Awaitable[PlaywrightPage]]

DEFAULT_VIEWPORT: ViewportSize = {"width": 1440, "height": 1024}


@asynccontextmanager
async def playwright_harness(
    *,
    headless: bool = True,
    cdp_endpoint: str | None = None,
    proxy: str | None = None,
    cookies: Sequence[SetCookieParam] = (),
    local_storage: Sequence[LocalStorage] = (),
    headers: dict[str, str] | None = None,
    stealth: bool = False,
    default_timeout: int = 30000,
    abort_unnecessary_requests: bool = True,
    user_agent: UserAgentFactory = random_user_agent,
    viewport: Optional[ViewportSize] = None,
    on_start: Optional[Callback] = None,
    on_end: Optional[Callback] = None,
    on_new_page: Optional[PageCallback] = None,
    browser_type: Optional[BrowserType] = None,
    enable_clipboard: bool = False,
    launch_args: Sequence[str] = (),
    extensions: Sequence[str] = (),
    **__: Any,
) -> AsyncGenerator[PageFactory, None]:
    """
    Context manager for Playwright. Starts a new browser, context, and page, and closes them when done.
    Also does some basic setup like setting the viewport, user agent, ignoring HTTPS errors,
    creation of HAR file, and stealth.
    """
    async with async_playwright() as p:
        extension_args = []
        browser_type = browser_type or "chromium"

        if extensions and browser_type == "chromium":
            extension_paths = ",".join(extensions)
            extension_args.extend(
                [
                    f"--disable-extensions-except={extension_paths}",
                    f"--load-extension={extension_paths}",
                ]
            )

        browser = await (
            p.chromium.connect_over_cdp(endpoint_url=cdp_endpoint)
            if cdp_endpoint
            else getattr(p, cast(str, browser_type)).launch(
                headless=headless,
                args=[
                    *launch_args,
                    *extension_args,
                    *(
                        # Disables navigator.webdriver showing up
                        ["--disable-blink-features=AutomationControlled"]
                        if browser_type == "chromium"
                        else []
                    ),
                    *(
                        # New chromium headless mode
                        ["--headless=new"] if headless else []
                    ),
                ],
            )
        )

        domain_storage = defaultdict(list)
        for item in local_storage:
            domain_storage[item["domain"]].append(item)

        # Create the storage state structure
        storage_state = {
            "origins": [
                {
                    "origin": f"https://{domain}",
                    "localStorage": [
                        {
                            "name": item["key"],
                            # Local storage only supports strings
                            "value": (
                                json.dumps(item["value"])
                                if isinstance(item["value"], (dict, list))
                                else str(item["value"])
                            ),
                        }
                        for item in items
                    ],
                }
                for domain, items in domain_storage.items()
            ]
        }

        ctx = await browser.new_context(
            viewport=viewport or DEFAULT_VIEWPORT,
            ignore_https_errors=True,
            user_agent=await compute_user_agent(user_agent),
            proxy=proxy_from_url(proxy) if proxy else None,
            permissions=["clipboard-read", "clipboard-write"]
            if enable_clipboard
            else None,
            storage_state=storage_state,
        )

        ctx.set_default_timeout(default_timeout)

        if headers:
            await ctx.set_extra_http_headers(headers)

        if cookies:
            await ctx.add_cookies(cookies)

        if abort_unnecessary_requests:
            await ctx.route("**/*", UnnecessaryResourceHandler().handle)

        async def page_factory(*_: Any, **__: Any) -> PlaywrightPage:
            page = await ctx.new_page()
            if on_new_page:
                await on_new_page(page)
            if stealth:
                await stealth_async(page)
            return page  # type: ignore

        try:
            if on_start:
                await on_start(ctx)
            yield page_factory
        finally:
            try:
                if on_end:
                    await on_end(ctx)
            finally:
                await ctx.close()
                await browser.close()



import ast
from pathlib import Path

import pytest

from harambe import SDK, meta


@SDK.scraper(domain="com.example", stage="listing")
async def example_scraper(sdk: SDK):
    pass


def test_find_scrapers():
    scrapers = meta.walk_package_for_decorators(Path(__file__))
    assert len(scrapers) == 1

    for scraper in scrapers:
        assert scraper["domain"]
        assert scraper["stage"]
        assert scraper["file_path"]
        assert scraper["function_name"]
        assert scraper["package"] == meta.url_to_package(scraper["domain"])


@pytest.mark.parametrize(
    "url, expected",
    [
        ("www.example.com", "com.example"),
        ("http://example.com", "com.example"),
        ("https://www.example.com", "com.example"),
        ("https://subdomain.example.com", "com.example.subdomain"),
        ("http://example.co.uk", "uk.co.example"),
    ],
)
def test_url_to_package(url, expected):
    assert meta.url_to_package(url) == expected


@pytest.mark.parametrize(
    "url, expected",
    [
        ("http://example.com", "example.com"),
        ("https://example.com", "example.com"),
        ("example.com", "example.com"),
        ("www.example.com", "example.com"),
        ("subdomain.example.com", "subdomain.example.com"),
        ("example.com:8000", "example.com"),
    ],
)
def test_url_to_netloc(url, expected):
    assert meta.url_to_netloc(url) == expected


@pytest.mark.parametrize(
    "node, expected",
    [
        (ast.Call(func=ast.Attribute(value=ast.Name(id="SDK"), attr="scraper")), True),
        (
            ast.Call(func=ast.Attribute(value=ast.Name(id="NotSDK"), attr="scraper")),
            False,
        ),
        (
            ast.Call(func=ast.Attribute(value=ast.Name(id="SDK"), attr="not_scraper")),
            False,
        ),
        (ast.Call(func=ast.Attribute(value=ast.Num(n=123), attr="scraper")), False),
        (ast.Num(n=123), False),
    ],
)
def test_is_sdk_scraper_decorator(node, expected):
    assert meta.is_sdk_scraper_decorator(node) == expected



from typing import Literal

from harambe.html_converter.html_to_markdown import HTMLToMarkdownConverter
from harambe.html_converter.html_to_text import HTMLToTextConverter
from harambe_core.errors import UnknownHTMLConverter
from markdownify import MarkdownConverter

HTMLConverterType = Literal["markdown", "text"]


def get_html_converter(
    html_converter_type: HTMLConverterType,
) -> MarkdownConverter:
    if html_converter_type == "markdown":
        return HTMLToMarkdownConverter()
    elif html_converter_type == "text":
        return HTMLToTextConverter()

    raise UnknownHTMLConverter(html_converter_type)
